/* F's Anti-Leechracer */
/* $Date: 2004-03-23 16:46:37 +0100 (Tue, 23 Mar 2004) $ $Rev: 1004 $ */

/* lots of includes */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include <unistd.h>
#include <sys/stat.h>
#include <time.h>

#include <regex.h>

/* our user config settings */
#include CONFIG_H

/* include a hex dump of the config file */
#include "inc-config.h"

/* some Limits */
#define MAXPATH       260

/* flag pattern matcher */

int flagsmatchpattern(const char *userflags, const char *flagspattern) {
     const char *flagptr;

     /* harhar, well this matches all flags from the pattern
	against the usersflags. all must be there. nice and short */

     while (*flagspattern) {
	  flagptr = userflags;
	  while(*flagptr) {
	       if (*flagspattern == *flagptr) break;
	       flagptr++;
	  }
	  if (!*flagptr) return 0;
	  flagspattern++;
     }
     return 1;
}

int flagsmatch(const char *userflags, const char *flagspatternlist) {
     char *flagspatternlisttoken = strdup(flagspatternlist);
     char *flagspattern;

     strcpy(flagspatternlisttoken, flagspatternlist);

     flagspattern = strtok(flagspatternlisttoken, ",");
     while(flagspattern) {
	  if (flagsmatchpattern(userflags, flagspattern)) {
	       free(flagspatternlisttoken);
	       return 1;
	  }
	  flagspattern = strtok(NULL, ",");
     }
     free(flagspatternlisttoken);
     return 0;
}

/* main program */

int main(int argc, char *argv[]) {
     const char *cmdstring, *loginname, *groupname;
     const char *envtest;

     if (argc == 2 && strcasecmp(argv[1],"dumpconfig") == 0) {
	  fwrite(confighdata, 1, sizeof confighdata, stdout);
	  return 0;
     }

     if (argc < 4) {
	  printf("Insufficient arguments. I should be called by glftpd as a cscript retr pre script.\n");
	  return 0;
     }

     cmdstring = argv[1];
     loginname = argv[2];
     groupname = argv[3];
 
     /***
	 First check if user has leech 
     ***/
#if ONLYLEECHCHECK == 1
     envtest = getenv("RATIO");
     if (!envtest) {
	  printf("You should be running this script from glftpd. It checks environment variables.\n");
	  return 1;
     }
     if (strcmp(envtest,"0") != 0) {
	  /* nope. not a leech user. ratio users are good */
	  return 0;
     }
#endif

     /***
	 check the user's name and group against the exclusive matches
     ***/
#ifdef USERIGNORE
     {
	  char *token;
	  char *userlist = strdup(USERIGNORE);

	  token = strtok(userlist, ",");
	  while(token) {
	       if (strcmp(token, loginname) == 0) {
		    /* Matches an exempted user */
		    free(userlist);
		    return 0;
	       }
	       token = strtok(NULL, ",");
	  }
	  free(userlist);
     }
#endif

#ifdef USERIGNORELIST
     {
	  FILE *f;
	  char line[256];

	  f = fopen(USERIGNORELIST, "r");
	  if (f) {
	       while( fgets(line, sizeof line, f) ) {
		    if (line[strlen(line)-1] == '\n') line[strlen(line)-1] = 0;
		    if (line[strlen(line)-1] == '\r') line[strlen(line)-1] = 0;

		    if (strcmp(line, loginname) == 0) {
			 /* matches a line in the file */
			 fclose(f);
			 return 0;
		    }
	       }
	       fclose(f);
	  }
	  /* else ignore error silently */
     }
#endif

#ifdef GROUPIGNORE
     {
	  char *token;
	  char *grouplist = strdup(GROUPIGNORE);

	  token = strtok(grouplist, ",");
	  while(token) {
	       if (strcasecmp(token, groupname) == 0) {
		    /* Matches an exempted group */
		    free(grouplist);
		    return 0;
	       }
	       token = strtok(NULL, ",");
	  }
	  free(grouplist);
     }
#endif

     /***
	 Now compare a users flags against include and exclude patterns
     ***/

     envtest = getenv("FLAGS");
     if (!envtest) {
	  printf("You should be running this script from glftpd. It checks environment variables.\n");
	  return 1;
     }

#ifdef FLAGSCHECK
     if (!flagsmatch(envtest, FLAGSCHECK)) {
	  /* none of the flag combos match the users flags, so he's free */
	  return 0;
     }
#endif

#ifdef FLAGSIGNORE
     if (flagsmatch(envtest, FLAGSIGNORE)) {
	  /* one of the flag combos matched the user. so he's exempted */
	  return 0;
     }
#endif

     {
	  /***
	      Compare cwd to path regexs
	  ***/

	  char cwd[MAXPATH];
	  
	  getcwd(cwd, sizeof cwd);

#ifdef PATHCHECK
	  {
	       regex_t reg;
	       regmatch_t regmatch[1];

	       if (regcomp(&reg, PATHCHECK, REG_EXTENDED) == 0) {
		    if (regexec(&reg, cwd, 0, regmatch, 0) != 0) {
			 /* path doesnt match -> let retr pass */
			 return 0;
		    }
	       }
	       else {
		    printf("550Cannot compile PATHCHECK regex pattern. Tell this the siteop.\r\n");
		    return 1;
	       }
	  }
#endif

#ifdef PATHIGNORE
	  {
	       regex_t reg;
	       regmatch_t regmatch[1];

	       if (regcomp(&reg, PATHIGNORE, REG_EXTENDED) == 0) {
		    if (regexec(&reg, cwd, 0, regmatch, 0) == 0) {
			 /* ignore path matches -> let retr pass */
			 return 0;
		    }
	       }
	       else {
		    printf("550Cannot compile PATHIGNORE regex pattern. Tell this the siteop.\r\n");
		    return 1;
	       }
	  }
#endif	  
     }
     
     {
	  /***
	      Parse out file name and run checks
	  ***/

	  if (strncasecmp(cmdstring, "RETR ", 5) == 0) {

	       const char *cpathptr;
	       char *pathstr;
	       struct stat st;
	       
	       cpathptr = cmdstring + 5;
	       while(*cpathptr == ' ' && *cpathptr != 0) cpathptr++;

	       if (*cpathptr == '/') {
		    char *newpath = malloc(strlen(HOMEDIRPATH) + strlen(cpathptr) + 1);
		    strcpy(newpath, HOMEDIRPATH);
		    strcat(newpath, cpathptr);

		    pathstr = newpath;
	       }
	       else {
		    pathstr = strdup(cpathptr);
	       }

	       if (stat(pathstr, &st) == 0) {
#if USEROWNEDFILES == 1
		    if (st.st_uid == getuid()) {
			 /* File uid matches running uid. */
			 printf("550File uid matches running uid");
			 return 0;
		    }
#endif		    
#if GROUPOWNEDFILES == 1
		    if (st.st_gid == getgid()) {
			 /* File gid matches running gid. */
			 return 0;
		    }
#endif

#ifdef FILEIGNORE
		    {
			 regex_t reg;
			 regmatch_t regmatch[1];

			 if (regcomp(&reg, FILEIGNORE, REG_EXTENDED) == 0) {
			      if (regexec(&reg, pathstr, 0, regmatch, 0) == 0) {
				   /* ignore file regex matches -> let retr pass */
				   return 0;
			      }
			 }
			 else {
			      printf("550Cannot compile FILEIGNORE regex pattern. Tell this the siteop.\r\n");
			      return 1;
			 }
		    }
#endif

#ifdef LEECHAGE
		    {
			 char timestr[32], *timeptr;
			 time_t tdiff = time(NULL) - st.st_mtime;

			 if (tdiff < LEECHAGE) {

			      tdiff = LEECHAGE - (time(NULL) - st.st_mtime);

			      timeptr = timestr;
			      *timeptr = 0;
			      if (tdiff > 3600) { timeptr += sprintf(timeptr, "%ldh ", tdiff / 3600); tdiff -= (int)(tdiff / 3600) * 3600; }
			      if (tdiff > 60) timeptr += sprintf(timeptr, "%ldm", tdiff / 60);
			      if (tdiff <= 60) { timeptr += sprintf(timeptr, "%lds", tdiff); }

			      printf("550Sorry, you are not allowed to leech this file yet. You need to wait %s before you may download.\r\n", timestr);
			      return 1;
			 }
			 /* else it's ok */
		    }
#endif
	       }
	       else {
		    free(pathstr);
		    /* File doesnt exist. */
		    printf("550File does not exist. If you are sure that it should exist tell this the siteop.\r\n");
		    return 1;
	       }
       
	       free(pathstr);
	  }
	  else {
	       printf("550Cannot parse command string. Tell this the siteop.\r\n");
	       return 1;
	  }
     }
     
     /* Nothing found */
     return 0;
}
